#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <math.h>

using namespace std;

// globale Variablen
const int N = 200;
const double h = 1./double(N-1);
const int MAX_ITER = N*N;
const double TOL = 1e-9;
const int M = 12;  //# Eigenwerten zu berechnen
const bool DEBUG = false;

// schreibe als ppm
void writePPM(const char* filename, const double data[][N]) {
	cout<<"Saving to file "<<filename<<endl;
	double min = data[0][0];
	double max = data[0][0];
  for( int i = 0; i < N; i++ )
    for( int j = 0; j < N; j++ ) {
		  if (data[i][j]<min) min=data[i][j];
		  if (data[i][j]>max) max=data[i][j];
    }

  ofstream out ( filename );
  out << "P3" << endl;
  out << N << " " << N << endl;
  out << 255 << endl;
  for( int i = 0; i < N; i++ )
    for( int j = 0; j < N; j++ ) {
    	int col = 255*(data[j][N-1-i]-min)/(max-min);
    	out << col << " " << col << " " << col << endl;
    }
}

// Skalar Produkt
double dot(const double chi[][N], const double x[][N], const double y[][N]) {
  double norm = 0.;
  for( int i = 0; i < N; i++ ) 
    for( int j = 0; j < N; j++ )
      if (chi[i][j]>0.5)
        norm += x[i][j] * y[i][j];
      return norm;
}

// Anwendung der Matrix L auf den Vecktor U,
// wobei L die finite Differenzen Diskretisierung des (negativen) Laplace-Operators ist.
// Keine Assemblierung der Matrix, sondern iterative Anwendung des Differenzensterns
//         |    -1     |
//  h^{-2} | -1  4  -1 |
//         |    -1     |
void applyMatrix(const double U[][N], const double chi[][N], double LU[][N] ) {
 
  const double h2 = h*h;

  for( int i = 0; i < N; i++ )
    for( int j = 0; j < N; j++ )
      if (chi[i][j]>0.5) {
       LU[i][j] = (4*U[i][j] - U[i-1][j] - U[i+1][j] - U[i][j-1] - U[i][j+1])/h2;
      } 
      else LU[i][j] = 0.;
}

double frand() {
  return rand()/float(RAND_MAX);
}


// Konjugierte Gradienten Methode zur Loesung des Systems Lx=b,
// wobei L die finite Differenzen Diskretisierung des (negativen) Laplace-Operators ist.
void applyCG(const double b[][N], double x[][N], const double chi[][N]) {
  static double r[N][N];
  static double p[N][N];
  static double Lp[N][N];

  // Erste Suchrichtung p = r_0
  // mit Residuum r_0 = b - L * x
  applyMatrix( x, chi, Lp );
  for( int i = 0; i < N; i++ ) 
    for( int j = 0; j < N; j++ )
      if (chi[i][j]>0.5) {
        r[i][j] = b[i][j] - Lp[i][j];
        p[i][j] = r[i][j];                   
      } 

  // berechne Norm des Residuums und Fehler
  double norm_rq = dot(chi,r,r);
  double error = sqrt( norm_rq ); 
  
  int iter = 1;

  // Iteriere bis die Fehlerschranke unterschritten bzw die 
  // maximale Iterationszahl erreicht ist.
  while ( ( iter < MAX_ITER) && ( error > TOL ) ) {
    // wende Matrix an
    applyMatrix( p, chi, Lp );
	
    // alpha = Norm(r)^2 / (p^T * L * p)
    double pLp = dot(chi,p,Lp);
    double alpha = norm_rq / pLp;       

    // Berechne Residuum r und Position x neu
    for( int i = 0; i < N; i++ ) 
      for( int j = 0; j < N; j++ )
        if (chi[i][j]>0.5) {
          x[i][j] += alpha * p[i][j];   // x = x + alpha * p
          r[i][j] -= alpha * Lp[i][j];  // r = r - alpha * L * p
        }      

    // berechne Norm des Residuums neu
    double norm_rq_alt = norm_rq;
    norm_rq = dot(chi,r,r);

    // Berechnung der Suchrichtung für den nächsten Schritt
    double beta = norm_rq / norm_rq_alt;
    // p_neu = beta * p_alt + r
    for( int i = 0; i < N; i++ ) 
      for( int j = 0; j < N; j++ )
        if (chi[i][j]>0.5)
          p[i][j] = beta * p[i][j] + r[i][j];

    // Aktualisierung des Fehlers
    error = sqrt( norm_rq );
    // Iterationszahl inkrementieren
    iter++;
  }
  
  // werte das Abbruchverhalten aus
  if ( error > TOL )
    cerr << "CG: Could not solve to required tolerance ( "<< TOL << " ) within " << MAX_ITER << " Iterations !!!" << endl;
  else
    if (DEBUG)
      cout << "CG: Solved to required tolerance ( "<< TOL << " ) within " << iter << " Iterations." << endl;
}

double normalize(const double chi[][N], double x[][N]) {
  double norm = sqrt(dot(chi,x,x));
  for( int i = 0; i < N; i++ ) 
    for( int j = 0; j < N; j++ )
      if (chi[i][j]>0.5)
        x[i][j] /= norm;
}

double copy(const double x[][N], double y[][N]) {
  for( int i = 0; i < N; i++ ) 
    for( int j = 0; j < N; j++ )
      y[i][j] = x[i][j];
}


//Inverse vector iteration
//X should have the K-1 previous eigenvectors
//calculate the Kth eigenvector and return the Kth eigenvalue
double invVektoriteration(const double chi[][N], double X[][N][N], const int K) {
  static double x0[N][N];
  static double x1[N][N];
  
  //TODO: setup random x0

  double mu0 = -1.0;
  double error = 2*TOL;   
  int iter = 1;
  while ( ( iter < MAX_ITER) && ( error > TOL ) ) {
    //TODO: subtract previous eigenvectors from xk(=x0)
    //TODO: solve Ax_{k+1} = x_k
    //TODO: update mu and normalize
    //TODO: update error
    if (DEBUG) cout<<error<<endl;
    mu0 = mu1;
    copy(x1,x0);
    iter++;
  }

  if ( error > TOL )
    cerr << "invVek: Could not solve to required tolerance ( "<< TOL << " ) within " << MAX_ITER << " Iterations !!!" << endl;
  else
    if (DEBUG)
      cout << "invVek: Solved to required tolerance ( "<< TOL << " ) within " << iter << " Iterations." << endl;

  //TODO: update the eigenvectors

  return 1./mu0;
}

// Hauptprogramm
int main () 
{
  // Definiere die Gebietsmaske chi
  static double chi[N][N];

  for( int i = 0; i < N; i++ ) {
    for( int j = 0; j < N; j++ ) {
      double x = i*h; //x,y in [0,1]
      double y = j*h;
      // das L-Gebiet ist gegeben durch \Omega = \{ (x,y) \in [0.1, 0.9]^2 : y > .5 falls x > .5 \}
      chi[i][j] = ((.1<x && x<.9 && .1<y && y<.9) && !(x>.5 && y<.5)) ? 1.0 : 0.0;
    }
  }
  
  // schreibe chi
  writePPM("chi.ppm",chi);

  //Eigenvektoren
  static double X[M][N][N];
  for (int m=0; m<M; m++) {
    double mu = invVektoriteration(chi, X, m);
    cout<<"Eigenvalue = "<<mu<<endl;
    stringstream fname;
    fname<<"x"<<m<<".ppm";
    writePPM(fname.str().c_str(),X[m]);
  }

  return 0;
}